@@ -14,12 +14,13 @@ use Sane;
 my $_POLL_INTERVAL;
 my $_self;
 my $buffer_size = ( 32 * 1024 );    # default size
-my ( $prog_name, $d, $logger );
+my ( $prog_name, $logger );
 
 sub setup {
- ( my $class, $prog_name, $d, $logger ) = @_;
- $_POLL_INTERVAL = 100;             # ms
+ ( my $class, $logger ) = @_;
+ $_POLL_INTERVAL = 100;                          # ms
  $_self          = {};
+ $prog_name      = Glib::get_application_name;
 
  $_self->{requests} = Thread::Queue->new;
  share $_self->{device_list};
@@ -142,8 +143,9 @@ sub find_scan_options {
   $finished_callback, $error_callback
  ) = @_;
 
- my $options : shared;
- my $sentinel = _enqueue_request( 'get-options', { options => \$options } );
+ my $option_array : shared;
+ my $sentinel =
+   _enqueue_request( 'get-options', { options => \$option_array } );
 
  my $started;
  _when_ready(
@@ -151,7 +153,7 @@ sub find_scan_options {
   sub {
    $started_callback->() unless ($started);
    if ( $_self->{status} == SANE_STATUS_GOOD ) {
-    $finished_callback->($options);
+    $finished_callback->($option_array);
    }
    else {
     $error_callback->( Sane::strstatus( $_self->{status} ) );
@@ -171,13 +173,13 @@ sub find_scan_options {
 sub set_option {
  my ( $class, %options ) = @_;
 
- my $options : shared;
+ my $option_array : shared;
  my $sentinel = _enqueue_request(
   'set-option',
   {
    index       => $options{index},
    value       => $options{value},
-   new_options => \$options
+   new_options => \$option_array
   }
  );
 
@@ -187,7 +189,7 @@ sub set_option {
   sub {
    $options{started_callback}->()
      if ( not $started and defined( $options{started_callback} ) );
-   $options{finished_callback}->($options)
+   $options{finished_callback}->($option_array)
      if ( defined $options{finished_callback} );
   },
   sub {
@@ -204,8 +206,8 @@ sub set_option {
 sub _new_page {
  my ( $dir, $format, $n ) = @_;
  my $path = sprintf $format, $n;
- return _enqueue_request( 'scan-page',
-  { path => File::Spec->catdir( $dir, $path ) } );
+ $path = File::Spec->catdir( $dir, $path ) if ( defined $dir );
+ return _enqueue_request( 'scan-page', { path => $path } );
 }
 
 sub scan_pages {
@@ -216,48 +218,32 @@ sub scan_pages {
  $_self->{scan_progress} = 0;
  my $sentinel = _new_page( $options{dir}, $options{format}, $options{start} );
 
- my $n2      = 1;
- my $npages2 = $options{npages};
-
+ my $n = 1;
  my $started;
  Glib::Timeout->add(
   $_POLL_INTERVAL,
   sub {
-   if ( not $$sentinel ) {
-    unless ($started) {
-     $options{started_callback}->() if ( defined $options{started_callback} );
-     $started = 1;
-    }
-    $options{running_callback}->( $_self->{scan_progress} )
-      if ( defined $options{running_callback} );
-    return Glib::SOURCE_CONTINUE;
-   }
-   else {
+   if ($$sentinel) {
 
     # Check status of scan
-    if ($_self->{status} == SANE_STATUS_GOOD
-     or $_self->{status} == SANE_STATUS_EOF )
-    {
-     $options{new_page_callback}->( $options{start} )
-       if ( defined $options{new_page_callback} );
-    }
-
-    # Stop the process unless everything OK and more scans required
-    unless (
-     ( $options{npages} == -1 or --$options{npages} )
+    $options{new_page_callback}->( $options{start} )
+      if (
+         defined( $options{new_page_callback} )
+     and not $_self->{abort_scan}
      and ( $_self->{status} == SANE_STATUS_GOOD
       or $_self->{status} == SANE_STATUS_EOF )
+      );
+
+    # Stop the process unless everything OK and more scans required
+    if (
+        $_self->{abort_scan}
+     or ( $options{npages} != -1 and not --$options{npages} )
+     or ( $_self->{status} != SANE_STATUS_GOOD
+      and $_self->{status} != SANE_STATUS_EOF )
       )
     {
      _enqueue_request('cancel');
-     if (
-         $_self->{status} == SANE_STATUS_GOOD
-      or $_self->{status} == SANE_STATUS_EOF
-      or ( $_self->{status} == SANE_STATUS_NO_DOCS
-       and $options{npages} < 1
-       and $n2 > 1 )
-       )
-     {
+     if ( _scanned_enough_pages( $options{npages}, $n ) ) {
       $options{finished_callback}->()
         if ( defined $options{finished_callback} );
      }
@@ -269,15 +255,48 @@ sub scan_pages {
     }
 
     $options{start} += $options{step};
-    $n2++;
+    $n++;
     $sentinel = _new_page( $options{dir}, $options{format}, $options{start} );
     return Glib::SOURCE_CONTINUE;
    }
+   else {
+    unless ($started) {
+     $options{started_callback}->() if ( defined $options{started_callback} );
+     $started = 1;
+    }
+    $options{running_callback}->( $_self->{scan_progress} )
+      if ( defined $options{running_callback} );
+    return Glib::SOURCE_CONTINUE;
+   }
   }
  );
  return;
 }
 
+sub _scanned_enough_pages {
+ my ( $ntodo, $ndone ) = @_;
+ return (
+       $_self->{status} == SANE_STATUS_GOOD
+    or $_self->{status} == SANE_STATUS_EOF
+    or ( $_self->{status} == SANE_STATUS_NO_DOCS
+   and $ntodo < 1
+   and $ndone > 1 )
+ );
+}
+
+# Flag the scan routine to abort
+
+sub cancel_scan {
+
+ # Empty process queue first to stop any new process from starting
+ $logger->info("Emptying process queue");
+ while ( $_self->{requests}->dequeue_nb ) { }
+
+ # Then send the thread a cancel signal
+ $_self->{abort_scan} = 1;
+ return;
+}
+
 sub _thread_main {
  my ($self) = @_;
 
@@ -324,6 +343,7 @@ sub _thread_open_device {
  undef $self->{device_handle} if ( defined( $self->{device_handle} ) );
 
  $self->{device_handle} = Sane::Device->open($device_name);
+ $logger->debug("opening device: $Sane::STATUS");
  if ( $Sane::STATUS != SANE_STATUS_GOOD ) {
   $logger->error("opening device: $Sane::STATUS");
   return;
@@ -367,7 +387,9 @@ sub _thread_set_option {
  $value = "$value";
 
  my $info = $self->{device_handle}->set_option( $index, $value );
- $logger->info("sane_set_option returned status $Sane::STATUS with info $info");
+ $logger->info(
+"sane_set_option $index to $value returned status $Sane::STATUS with info $info"
+ );
 
  # FIXME: This duplicates _thread_get_options.
  if ( $info & SANE_INFO_RELOAD_OPTIONS ) {
@@ -447,79 +469,19 @@ sub _thread_scan_page_to_fh {
     goto cleanup;
    }
 
-   if ($first_frame) {
-    if ( $parm->{lines} >= 0 ) {
-     $logger->info(
-      sprintf "$prog_name: scanning image of size %dx%d pixels at "
-        . "%d bits/pixel",
-      $parm->{pixels_per_line},
-      $parm->{lines},
-      8 * $parm->{bytes_per_line} / $parm->{pixels_per_line}
-     );
-    }
-    else {
-     $logger->info(
-      sprintf "$prog_name: scanning image %d pixels wide and "
-        . "variable height at %d bits/pixel",
-      $parm->{pixels_per_line},
-      8 * $parm->{bytes_per_line} / $parm->{pixels_per_line}
-     );
-    }
-
-    $logger->info(
-     sprintf "$prog_name: acquiring %s frame",
-     $parm->{format} <= SANE_FRAME_BLUE
-     ? $format_name[ $parm->{format} ]
-     : "Unknown"
-    );
-   }
-
-   if ($first_frame) {
-    if ($parm->{format} == SANE_FRAME_RED
-     or $parm->{format} == SANE_FRAME_GREEN
-     or $parm->{format} == SANE_FRAME_BLUE )
-    {
-     die "Red/Green/Blue frames require depth=8\n"
-       unless ( $parm->{depth} == 8 );
-     $must_buffer = 1;
-     $offset      = $parm->{format} - SANE_FRAME_RED;
-    }
-    elsif ( $parm->{format} == SANE_FRAME_RGB ) {
-     die "RGB frames require depth=8 or 16\n"
-       unless ( ( $parm->{depth} == 8 ) or ( $parm->{depth} == 16 ) );
-    }
-    if ($parm->{format} == SANE_FRAME_RGB
-     or $parm->{format} == SANE_FRAME_GRAY )
-    {
-     die "Valid depths are 1, 8 or 16\n"
-       unless ( ( $parm->{depth} == 1 )
-      or ( $parm->{depth} == 8 )
-      or ( $parm->{depth} == 16 ) );
-     if ( $parm->{lines} < 0 ) {
-      $must_buffer = 1;
-      $offset      = 0;
-     }
-     else {
-      _thread_write_pnm_header( $fh, $parm->{format}, $parm->{pixels_per_line},
-       $parm->{lines}, $parm->{depth} );
-     }
-    }
-   }
-   else {
-    die "Encountered unknown format\n"
-      if ( $parm->{format} < SANE_FRAME_RED
-     or $parm->{format} > SANE_FRAME_BLUE );
-    $offset = $parm->{format} - SANE_FRAME_RED;
-    $image{x} = $image{y} = 0;
-   }
-   my $hundred_percent = $parm->{bytes_per_line} * $parm->{lines} * (
-    ( $parm->{format} == SANE_FRAME_RGB or $parm->{format} == SANE_FRAME_GRAY )
-    ? 1
-    : 3
-   );
+   _log_frame_info( $first_frame, $parm, \@format_name );
+   ( $must_buffer, $offset ) = _initialise_scan( $fh, $first_frame, $parm );
+   my $hundred_percent = _scan_data_size($parm);
 
    while (1) {
-    $device->cancel if ( $_self->{abort_scan} );
+
+    # Pick up flag from cancel_scan()
+    if ( $_self->{abort_scan} ) {
+     $device->cancel;
+     $logger->info('Scan cancelled');
+     return;
+    }
+
     my ( $buffer, $len ) = $device->read($buffer_size);
     $total_bytes += $len;
     my $progr = $total_bytes / $hundred_percent;
@@ -538,31 +500,9 @@ sub _thread_scan_page_to_fh {
     }
 
     if ($must_buffer) {
-
-     # We're either scanning a multi-frame image or the
-     # scanner doesn't know what the eventual image height
-     # will be (common for hand-held scanners).  In either
-     # case, we need to buffer all data before we can write
-     # the image
-     if ($parm->{format} == SANE_FRAME_RED
-      or $parm->{format} == SANE_FRAME_GREEN
-      or $parm->{format} == SANE_FRAME_BLUE )
-     {
-      for ( my $i = 0 ; $i < $len ; ++$i ) {
-       $image{data}[ $offset + 3 * $i ] = substr( $buffer, $i, 1 );
-      }
-      $offset += 3 * $len;
-     }
-     elsif ( $parm->{format} == SANE_FRAME_RGB
-      or $parm->{format} == SANE_FRAME_GRAY )
-     {
-      for ( my $i = 0 ; $i < $len ; ++$i ) {
-       $image{data}[ $offset + $i ] = substr( $buffer, $i, 1 );
-      }
-      $offset += $len;
-     }
+     $offset = _buffer_scan( $offset, $parm, \%image, $len, $buffer );
     }
-    else {    # ! must_buffer
+    else {
      print $fh $buffer;
     }
    }
@@ -570,21 +510,7 @@ sub _thread_scan_page_to_fh {
   } while ( !$parm->{last_frame} );
  }
 
- if ($must_buffer) {
-  if ( $parm->{lines} > 0 ) {
-   $image{height} = $parm->{lines};
-  }
-  else {
-   $image{height} = @{ $image{data} } / $parm->{pixels_per_line};
-   $image{height} /= 3
-     if ( $parm->{format} == SANE_FRAME_RED
-    or $parm->{format} == SANE_FRAME_GREEN
-    or $parm->{format} == SANE_FRAME_BLUE );
-  }
-  _thread_write_pnm_header( $fh, $parm->{format}, $parm->{pixels_per_line},
-   $image{height}, $parm->{depth} );
-  for ( @{ $image{data} } ) { print $fh; }
- }
+ _write_buffer_to_fh( $fh, $parm, \%image ) if ($must_buffer);
 
 cleanup:
  my $expected_bytes = $parm->{bytes_per_line} * $parm->{lines} * (
@@ -608,7 +534,12 @@ cleanup:
 sub _thread_scan_page {
  my ( $self, $path ) = @_;
 
+ unless ( defined $self->{device_handle} ) {
+  $logger->info("$prog_name: must open device before starting scan");
+  return;
+ }
  $self->{device_handle}->start;
+
  $self->{status} = $Sane::STATUS + 0;
  if ( $Sane::STATUS != SANE_STATUS_GOOD ) {
   $logger->info("$prog_name: sane_start: $Sane::STATUS");
@@ -641,7 +572,132 @@ sub _thread_scan_page {
 
 sub _thread_cancel {
  my ($self) = @_;
- $self->{device_handle}->cancel;
+ $self->{device_handle}->cancel if ( defined $self->{device_handle} );
+ return;
+}
+
+sub _log_frame_info {
+ my ( $first_frame, $parm, $format_name ) = @_;
+ if ($first_frame) {
+  if ( $parm->{lines} >= 0 ) {
+   $logger->info(
+    sprintf "$prog_name: scanning image of size %dx%d pixels at "
+      . "%d bits/pixel",
+    $parm->{pixels_per_line},
+    $parm->{lines},
+    8 * $parm->{bytes_per_line} / $parm->{pixels_per_line}
+   );
+  }
+  else {
+   $logger->info(
+    sprintf "$prog_name: scanning image %d pixels wide and "
+      . "variable height at %d bits/pixel",
+    $parm->{pixels_per_line},
+    8 * $parm->{bytes_per_line} / $parm->{pixels_per_line}
+   );
+  }
+
+  $logger->info(
+   sprintf "$prog_name: acquiring %s frame",
+   $parm->{format} <= SANE_FRAME_BLUE
+   ? $format_name->[ $parm->{format} ]
+   : "Unknown"
+  );
+ }
+ return;
+}
+
+sub _initialise_scan {
+ my ( $fh, $first_frame, $parm ) = @_;
+ my ( $must_buffer, $offset );
+ if ($first_frame) {
+  if ($parm->{format} == SANE_FRAME_RED
+   or $parm->{format} == SANE_FRAME_GREEN
+   or $parm->{format} == SANE_FRAME_BLUE )
+  {
+   die "Red/Green/Blue frames require depth=8\n"
+     unless ( $parm->{depth} == 8 );
+   $must_buffer = 1;
+   $offset      = $parm->{format} - SANE_FRAME_RED;
+  }
+  elsif ( $parm->{format} == SANE_FRAME_RGB ) {
+   die "RGB frames require depth=8 or 16\n"
+     unless ( ( $parm->{depth} == 8 ) or ( $parm->{depth} == 16 ) );
+  }
+  if ($parm->{format} == SANE_FRAME_RGB
+   or $parm->{format} == SANE_FRAME_GRAY )
+  {
+   die "Valid depths are 1, 8 or 16\n"
+     unless ( ( $parm->{depth} == 1 )
+    or ( $parm->{depth} == 8 )
+    or ( $parm->{depth} == 16 ) );
+   if ( $parm->{lines} < 0 ) {
+    $must_buffer = 1;
+    $offset      = 0;
+   }
+   else {
+    _thread_write_pnm_header( $fh, $parm->{format}, $parm->{pixels_per_line},
+     $parm->{lines}, $parm->{depth} );
+   }
+  }
+ }
+ else {
+  die "Encountered unknown format\n"
+    if ( $parm->{format} < SANE_FRAME_RED
+   or $parm->{format} > SANE_FRAME_BLUE );
+  $offset = $parm->{format} - SANE_FRAME_RED;
+ }
+ return ( $must_buffer, $offset );
+}
+
+# Return size of final scan (ignoring header)
+
+sub _scan_data_size {
+ my ($parm) = @_;
+ return $parm->{bytes_per_line} * $parm->{lines} * (
+  ( $parm->{format} == SANE_FRAME_RGB or $parm->{format} == SANE_FRAME_GRAY )
+  ? 1
+  : 3
+ );
+}
+
+# We're either scanning a multi-frame image or the
+# scanner doesn't know what the eventual image height
+# will be (common for hand-held scanners).  In either
+# case, we need to buffer all data before we can write
+# the header
+
+sub _buffer_scan {
+ my ( $offset, $parm, $image, $len, $buffer ) = @_;
+
+ # $parm->{format} == SANE_FRAME_RED or SANE_FRAME_GREEN or SANE_FRAME_BLUE
+ my $number_frames = 3;
+ $number_frames = 1
+   if ( $parm->{format} == SANE_FRAME_RGB
+  or $parm->{format} == SANE_FRAME_GRAY );
+
+ for ( my $i = 0 ; $i < $len ; ++$i ) {
+  $image->{data}[ $offset + $number_frames * $i ] = substr( $buffer, $i, 1 );
+ }
+ $offset += $number_frames * $len;
+ return $offset;
+}
+
+sub _write_buffer_to_fh {
+ my ( $fh, $parm, $image ) = @_;
+ if ( $parm->{lines} > 0 ) {
+  $image->{height} = $parm->{lines};
+ }
+ else {
+  $image->{height} = @{ $image->{data} } / $parm->{pixels_per_line};
+  $image->{height} /= 3
+    if ( $parm->{format} == SANE_FRAME_RED
+   or $parm->{format} == SANE_FRAME_GREEN
+   or $parm->{format} == SANE_FRAME_BLUE );
+ }
+ _thread_write_pnm_header( $fh, $parm->{format}, $parm->{pixels_per_line},
+  $image->{height}, $parm->{depth} );
+ for ( @{ $image->{data} } ) { print $fh; }
  return;
 }
 
